-- Hypatia
-- version 4.5

-- global procedure makenumeric - converts numeric elements of input stack to numbers
-- function need - checks whether sufficient number of arguments on stack for operator
-- procedure done - deletes all elements but the last one left of DONE pseudo operator from stack
-- procedure restorehead - after n-argument operator, restores elements of stack that were cut off by |, if any
-- global procedure calculate - parses stack and makes operators perform calculations

include pmaths.e

integer ix -- current index in stack
integer first -- first operator in line - if 1, then need() can provide $ as first argument
integer flag_im -- show intermediate results (debug mode)
atom x -- result of single operation
atom y -- temp variable
atom a, b -- stack[ix-2], stack[ix-1], assigned by need()
object head -- head of stack, cut off by | pseudo operator, to be restored by n-argument operator
object opc -- operator

global procedure makenumeric()
	-- converts numeric strings into their values
	-- replaces constants and variables with their numeric values
	-- non-numeric elements (which should be operators) are changed into sequences, with " " as[1]
	-- variable names and commands are converted to lower case
	-- checks for # ? & at line end and sets flag_silent, flag_im or flag_acc if found
	-- valid combinations: ? &, & ?, ? #

	object item
	object v
	integer ok

	flag_silent = 0
	if equal (stack[$], "#") then            -- checks for # at line end
		flag_silent = 1
		stack = stack[1..$-1]
	end if

	flag_im = 0
	if equal (stack[$], "?") then            -- checks for ? at line end
		flag_im = 1
		stack = stack[1..$-1]
	end if

	flag_acc = 0
	if equal (stack[$], "&") or equal(stack[$], "&&") then            -- checks for & or && at line end, && = new line accumulation mode
		if flag_silent then
			err = 1
			errtxt = "accumulation mode and silent mode contradict each other"
			return
		else
			if equal (stack[$], "&") then
				flag_acc = 1
			else
				flag_acc = 2
			end if
			stack = stack[1..$-1]
			if flag_im = 0 then                    -- checks for ? at line end if order is ? & or ? &&
				if equal (stack[$], "?") then
					flag_im = 1
					stack = stack[1..$-1]
				end if
			end if
		end if
	end if

	for j = 1 to length(stack) do

		item = lower(stack[j])
		v = readnumber(item)

		if atom(v) then
			stack[j] = v

		-- define constants
		-- the constant names must also be checked by the command line parser to determine CLC mode
		elsif equal(item, "pi") then stack[j] = PI
		elsif equal(item, "tau") then stack[j] = PI * 2
		elsif equal(item, "e") then stack[j] = EULER
		elsif equal(item, "phi") then stack[j] = 1.618033988749894848
		elsif equal(item, "rand") then stack[j] = rnd()
		-- end of constants

		elsif equal(item, "i") then
			if flag_do then               -- do loop: I = loop index
				stack[j] = i_repeat
			else                          -- repeat loop: I = loop index + 1
				stack[j] = i_repeat + 1
			end if

		elsif equal(item, "isloop") then
			stack[j] = sign(num_repeat)

		elsif equal(item, "passes") then
			if num_repeat > 0 then
				stack[j] = num_repeat
			else
				stack[j] = 1
			end if
		
		elsif equal(item, "time") then
			if sign(num_repeat) then
				stack[j] = time() - time_start
			else
				stack[j] = 0
			end if

		elsif equal(item, "dup") then
			stack[j] = var[1][2]
			for jj = j-1 to 1 by -1 do
				if atom(stack[jj]) then
					stack[j] = stack[jj]
					exit
				end if
			end for

		elsif item[1] = '$' then
			ok = 0
			for jj = 1 to length(var) do
				if equal(item, var[jj][1]) then
					stack[j] = var[jj][2]
					ok = 1
					exit
				end if
			end for
			if ok = 0 then
				err = 1
				errtxt = "variable " & stack [j] & " not found"
			end if
		else
			stack[j] = append(" ", stack[j])
		end if
		if err then exit end if
	end for
end procedure


function need(integer n)
	-- checks whether enough arguments requested by operator are available
	-- for first operator in line, if 1 argument missing prepends value of $ to stack
	-- returns 0 if failed, 1 if ok

	if ix > n then
		if ix != 1 and find(opc, {"lakh", "million", "crore"}) = 0 then
			first = 0
		end if
		if ix > 2 then a = stack[ix-2] end if
		b = stack[ix-1]
		return 1
	else
		if first = 1 and ix = n then
			if mode_autoinc then
				stack = prepend(stack, var[1][2])
				ix = ix + 1
				if ix > 2 then a = stack[ix-2] end if
				if ix > 1 then b = stack[ix-1] end if
				first = 0
				return 1
			else
				err = 1
				errtxt = "argument missing for first operator, auto include $ is OFF"
				return 0
			end if
		else
			err = 1
			errtxt = "argument(s) missing for operator " & upper(stack[ix][2])
			return 0
		end if
	end if
end function


procedure done(integer n)
	-- deletes n elements before operator, and replaces operator with x
	integer a
	if abs(x) > 1.001E1000 then
		err = 1
		errtxt = "operator " & upper(opc) & " result > 1e1000"
	else
		if abs(x) < 1e-1000 then
			x = 0
		elsif mode_integerbias then
			x = snapinteger(x)
		end if
		if flag_im then disp("[ op: " & stack[ix][2] & " | result: " & sprint15(x, fmt_digits) & " | position: " & sprint(ix - n) & " ]") end if
		stack[ix] = x
		a = ix - n - 1
		if a = 0 then
			stack = stack[ix..$]
		else
			stack = stack[1..a] & stack[ix..$]
		end if
		ix = ix - n
	end if
end procedure


procedure restorehead()
	-- after n-argument operator, restores head of stack that was cut off by |
	if length(head) then
		stack = head & stack
		ix = ix + length(head)
		head = ""
	end if
end procedure


global procedure calculate()
	atom flag_fw  -- first word in line, 1 = yes
	atom flag_whisk -- 1 after whisk operator
	atom whisk_a, whisk_b -- whisk cut/paste variables
	flag_whisk = 0
	flag_fw = 1
	head = ""
	ix = 0
	first = 1 -- indicates first operation in line not yet executed, allows last result to be used as first argument (set 0 by n-argument operators, whisk, and done)

	while ix < length(stack) do
		ix = ix + 1
		if atom(stack[ix]) = 0 then
			opc = lower(stack[ix][2])
			err = 0
			x = 0

			-- operators with 1 argument

			if equal(opc, "+-") then
				if need(1) then
					x = -b
					done(1)
				end if
			elsif equal(opc, "rcp") then
				if need(1) then
					if b = 0 then
						err = 1
						errtxt = "reciprocal value of 0"
					else
						x = 1 / b
						done(1)
					end if
				end if
			elsif equal(opc, "sign") then
				if need(1) then
					b = round0(b)
					x = sign(b)
					done(1)
				end if
			elsif equal(opc, "isnot0") then
				if need(1) then
					b = round0(b)
					if b = 0 then x = 0 else x = 1 end if
					done(1)
				end if
			elsif equal(opc, "is+") then
				if need(1) then
					b = round0(b)
					if b > 0 then x = 1 else x = 0 end if
					done(1)
				end if
			elsif equal(opc, "is+0") or equal(opc, "is0+") then
				if need(1) then
					b = round0(b)
					if b < 0 then x = 0 else x = 1 end if
					done(1)
				end if
			elsif equal(opc, "is0") then
				if need(1) then
					b = round0(b)
					if b = 0 then x = 1 else x = 0 end if
					done(1)
				end if
			elsif equal(opc, "is-") then
				if need(1) then
					b = round0(b)
					if b < 0 then x = 1 else x = 0 end if
					done(1)
				end if
			elsif equal(opc, "is-0") or equal(opc, "is0-") then
				if need(1) then
					b = round0(b)
					if b > 0 then x = 0 else x = 1 end if
					done(1)
				end if
			elsif equal(opc, "abs") then
				if need(1) then
					x = abs(b)
					done(1)
				end if
			elsif equal(opc, "int") then
				if need(1) then
					x = abs(b)
					if x >= 1E18 then
						err = 1
						errtxt = "INT argument must be < 1e18"
					else
						x = floor(x)
						if b < 0 then x = -x end if
						done(1)
					end if
				end if
			elsif equal(opc, "frac") then
				if need(1) then
					x = abs(b)
					if x >= 1E15 then
						err = 1
						errtxt = "FRAC argument must be < 1e15"
					else
						integer mx
						mx = magntd(x)
						x = x - floor(x)
						x = rround(x, 17 - mx)
						if x > 1 - 1E-15 then
							x = 1 - 1E-15
						end if
						if b < 0 then x = -x end if
						done(1)
					end if
				end if
			elsif equal(opc, "isprime") then
				if need(1) then
					if b < 0 then
						err = 1
						errtxt = "ISPRIME argument < 0"
					elsif abs(b) >= 1E18 then
						err = 1
						errtxt = "ISPRIME argument >= 1e18"
					elsif isnat(b) = 0 then
						err = 1
						errtxt = "ISPRIME argument is not integer"
					else
						x = is_prime(b)
						done(1)
					end if
				end if
			elsif equal(opc, "isposint") then
				if need(1) then
					if b > 0 then
						if b >= 1E18 then
							err = 1
							errtxt = "ISPOSINT argument >= 1e18"
						else
							x = integer(b)
						end if
					else
						x = 0
					end if
					done(1)
				end if
			elsif equal(opc, "digits") then
				if need(1) then
					x = magntd(b) + 1
					done(1)
				end if
			elsif equal(opc, "sq") then
				if need(1) then
					x = b * b
					done(1)
				end if
			elsif equal(opc, "sqrt") then
				if need(1) then
					if b < 0 then
						err = 1
						errtxt = "square root of negative number"
					else
						x = sqrt(b)
						done(1)
					end if
				end if
			elsif equal(opc, "cube") then
				if need(1) then
					x = b * b * b
					done(1)
				end if
			elsif equal(opc, "cbrt") then
				if need(1) then
					if b < 0 then
						x = -exp(log(-b)/3)
					elsif b = 0 then
						x = 0
					else
						x = exp(log(b)/3)
					end if
					done(1)
				end if
			elsif equal(opc, "ln") then
				if need(1) then
					if b <= 0 then
						err = 1
						errtxt = "logarithm of 0 or negative number"
					else
						x = log(b)
						done(1)
					end if
				end if
			elsif equal(opc, "log10") then
				if need(1) then
					if b <= 0 then
						err = 1
						errtxt = "logarithm of 0 or negative number"
					else
						x = log10(b)
						done(1)
					end if
				end if
			elsif equal(opc, "log2") then
				if need(1) then
					if b <= 0 then
						err = 1
						errtxt = "logarithm of 0 or negative number"
					else
						x = log2(b)
						done(1)
					end if
				end if
			elsif equal(opc, "exp") then
				if need(1) then
					if b <= 2303 then
						x = exp(b)
						done(1)
					else
						err = 1
						errtxt = "operator EXP result > 1e1000"
					end if
				end if
			elsif equal(opc, "exp10") then
				if need(1) then
					if b <= 1000 then
						x = apower(10, b)
						done(1)
					else
						err = 1
						errtxt = "operator EXP10 result > 1e1000"
					end if
				end if
			elsif equal(opc, "exp2") then
				if need(1) then
					if b <= 3322 then
						x = apower(2, b)
						done(1)
					else
						err = 1
						errtxt = "operator EXP2 result > 1e1000"
					end if
				end if
			elsif equal(opc, "!") then
				if need(1) then
					if isnat(b) then
						if b > 449 then
							err = 1
							errtxt = "factorial of number > 449"
						else
							x = 1
							for j = 1 to b do
								x = x * j
							end for
							done(1)
						end if
					else
						err = 1
						if b < 0 then
							errtxt = "factorial of negative number"
						else
							errtxt = "factorial of non-integer number"
						end if
					end if
				end if
			elsif equal(opc, "rad") then
				if need(1) then
					x = b * PI / 180
					done(1)
				end if
			elsif equal(opc, "deg") then
				if need(1) then
					x = b * 180 / PI
					done(1)
				end if
			elsif equal(opc, "sin") then
				if need(1) then
					if mode_deg then b = b * PI / 180 end if
					x = sin(b)
					if abs(x) < 1E-18 then x = 0 end if
					done(1)
				end if
			elsif equal(opc, "cos") then
				if need(1) then
					if mode_deg then b = b * PI / 180 end if
					x = cos(b)
					if abs(x) < 1E-18 then x = 0 end if
					done(1)
				end if
			elsif equal(opc, "tan") then
				if need(1) then
					if mode_deg then b = b * PI / 180 end if
					if abs(b) > 1.5707963267948966 then
						err = 1
						errtxt = "TAN argument must be < 90 and > -90 degrees"
					else
						x = tan(b)
						done(1)
					end if
				end if
			elsif equal(opc, "asin") then
				if need(1) then
					if b > 1 or b < -1 then
						err = 1
						errtxt = "Arcsine argument must be <= 1 and >= -1"
					else
						x = arcsin(b)
						if mode_deg then x = x * 180 / PI end if
						done(1)
					end if
				end if
			elsif equal(opc, "acos") then
				if need(1) then
					if b > 1 or b < -1 then
						err = 1
						errtxt = "Arccosine argument must be <= 1 and >= -1"
					else
						x = arccos(b)
						if mode_deg then x = x * 180 / PI end if
						done(1)
					end if
				end if
			elsif equal(opc, "atan") then
				if need(1) then
					x = arctan(b)
					if mode_deg then x = x * 180 / PI end if
					done(1)
				end if

			-- lakh, million, crore
			
			elsif equal(opc, "lakh") then
				if need(1) then
					x = b * 1E5
					done(1)
				end if
			elsif equal(opc, "million") then
				if need(1) then
					x = b * 1E6
					done(1)
				end if
			elsif equal(opc, "crore") then
				if need(1) then
					x = b * 1E7
					done(1)
				end if

			-- measurement unit conversion operators (1 argument)

			elsif equal(opc, ":f") then
				if need(1) then
					x = b * 1.8 + 32
					done(1)
				end if
			elsif equal(opc, ":c") then
				if need(1) then
					x = (b - 32) / 1.8
					done(1)
				end if
			elsif equal(opc, ":mi") then
				if need(1) then
					x = b / 1.609344
					done(1)
				end if
			elsif equal(opc, ":km") then
				if need(1) then
					x = b * 1.609344
					done(1)
				end if
			elsif equal(opc, ":nautmi") then
				if need(1) then
					x = b / 1.852
					done(1)
				end if
			elsif equal(opc, ":nautkm") then
				if need(1) then
					x = b * 1.852
					done(1)
				end if
			elsif equal(opc, ":in") then
				if need(1) then
					x = b / 2.54
					done(1)
				end if
			elsif equal(opc, ":cm") then
				if need(1) then
					x = b * 2.54
					done(1)
				end if
			elsif equal(opc, ":ft") then
				if need(1) then
					x = b / 0.3048
					done(1)
				end if
			elsif equal(opc, ":m") then
				if need(1) then
					x = b * 0.3048
					done(1)
				end if
			elsif equal(opc, ":lb") then
				if need(1) then
					x = b / 0.45359237
					done(1)
				end if
			elsif equal(opc, ":kg") then
				if need(1) then
					x = b * 0.45359237
					done(1)
				end if
			elsif equal(opc, ":oz") then
				if need(1) then
					x = b / 28.349523125
					done(1)
				end if
			elsif equal(opc, ":g") then
				if need(1) then
					x = b * 28.349523125
					done(1)
				end if
			elsif equal(opc, ":toz") then
				if need(1) then
					x = b / 31.1034768
					done(1)
				end if
			elsif equal(opc, ":tg") then
				if need(1) then
					x = b * 31.1034768
					done(1)
				end if
			elsif equal(opc, ":gal") then
				if need(1) then
					x = b / 3.785411784
					done(1)
				end if
			elsif equal(opc, ":l") then
				if need(1) then
					x = b * 3.785411784
					done(1)
				end if
			elsif equal(opc, ":mpg") then
				if need(1) then
					x = 235.2145833 / b
					done(1)
				end if

			-- operators with 2 arguments

			elsif equal(opc, "+") then
				if need(2) then
					x = a + b
					done(2)
				end if
			elsif equal(opc, "-") then
				if need(2) then
					x = a - b
					done(2)
				end if
			elsif equal(opc, "*") then
				if need(2) then
					x = a * b
					done(2)
				end if
			elsif equal(opc, "/") then
				if need(2) then
					if b = 0 then
						err = 1
						errtxt = "division by 0"
					else
						x = a / b
						done(2)
					end if
				end if
			elsif equal(opc, "//") then
				if need(2) then
					if b = 0 then
						x = a
					else
						x = (a - b) / b
					end if
					if abs(x) < def_zero then x = 0 end if
					done(2)
				end if
			elsif equal(opc, "\\") then
				if need(2) then
					if b = 0 then
						err = 1
						errtxt = "division by 0"
					else
						x = floor(abs(a) / abs(b))
						x = x * sign(a) * sign(b)
						done(2)
					end if
				end if
			elsif equal(opc, "equal") or equal(opc, "==") then
				if need(2) then
					x = round0(a - b)
					if x = 0 then x = 1 else x = 0 end if
					done(2)
				end if
			elsif equal(opc, "unequal") or equal(opc, "<>") then
				if need(2) then
					x = round0(a - b)
					if x != 0 then x = 1 end if
					done(2)
				end if
			elsif equal(opc, "<<") then
				if need(2) then
					x = round0(a - b)
					if x < 0 then x = 1 else x = 0 end if
					done(2)
				end if
			elsif equal(opc, "<=") then
				if need(2) then
					x = round0(a - b)
					if x <= 0 then x = 1 else x = 0 end if
					done(2)
				end if
			elsif equal(opc, ">>") then
				if need(2) then
					x = round0(a - b)
					if x > 0 then x = 1 else x = 0 end if
					done(2)
				end if
			elsif equal(opc, ">=") then
				if need(2) then
					x = round0(a - b)
					if x >= 0 then x = 1 else x = 0 end if
					done(2)
				end if
			elsif equal(opc, "multiple") then
				if need(2) then
					if abs(a) >= 1E18 or abs(b) >= 1E18 then
						err = 1
						errtxt = "arguments for multiple must be < 1e18"
					elsif b = 0 then
						err = 1
						errtxt = "multiple of 0"
					elsif not integer(a) or not integer(b) then
						err = 1
						errtxt = "non-integer argument(s) for multiple"
					else
						x = a / b
						if x < 0 or integer(x) = 0 then x = 0 end if
					end if
					done(2)
				end if
			elsif equal(opc, "round") then
				if need(2) then
					if isnat(abs(b)) = 0 then
						err = 1
						errtxt = "rounding to non-integer number of decimals"
					elsif b < -9 or b > 15 then
						err = 1
						errtxt = "digits for rounding must be -9 to 15"
					else
						x = rround(a, b)
						done(2)
					end if
				end if
			elsif equal(opc, "rounds") then
				if need(2) then
					if isnat(abs(b)) = 0 then
						err = 1
						errtxt = "rounding to non-integer number of digits"
					elsif b < 1 or b > 17 then
						err = 1
						errtxt = "sign. digits for rounding must be 1 to 17"
					else
						x = roundsig(a, b)
						done(2)
					end if
				end if
			elsif equal(opc, "gate") then
				if need(2) then
					if b = 0 then
						if zerogate then
							b = zerogate
						else
							b = def_zero
						end if
					end if
					if b < 0 then
						err = 1
						errtxt = "threshold value is negative"
					else
						if abs(a) < b then
							x = 0
						else
							x = a
						end if
					end if
					done(2)
				end if
			elsif equal(opc, "uplimit") then
				if need(2) then
					if a > b then
						x = b
					else
						x = a
					end if
					done(2)
				end if
			elsif equal(opc, "lolimit") then
				if need(2) then
					if a < b then
						x = b
					else
						x = a
					end if
					done(2)
				end if
			elsif equal(opc, "abslimit") then
				if need(2) then
					if b <= 0 then
						err = 1
						errtxt = "threshold value is 0 or negative"
					else		
						if abs(a) > b then
							if a < 0 then
								x = -b
							else
								x = b
							end if
						else
							x = a
						end if
						done(2)
					end if
				end if
			elsif equal(opc, "%") then
				if need(2) then
					x = a * b / 100
					done(2)
				end if
			elsif equal(opc, "%+") then
				if need(2) then
					x = a * (1 + b/100)
					done(2)
				end if
			elsif equal(opc, "%-") then
				if need(2) then
					x = a * (1 - b/100)
					done(2)
				end if
			elsif equal(opc, "%?") then
				if need(2) then
					if a = 0 then
						err = 1
						errtxt = "% division by 0"
					else
						x = b / a *100
						done(2)
					end if
				end if
			elsif equal(opc, "%n") then
				if need(2) then
					if b = -100 then
						err = 1
						errtxt = "% division by 0"
					else
						x = a / (1 + b/100)
						done(2)
					end if
				end if
			elsif equal(opc, "%a") then
				if need(2) then
					if b = -100 then
						err = 1
						errtxt = "% division by 0"
					else
						x = a - a / (1 + b/100)
						done(2)
					end if
				end if
			elsif equal(opc, "imod") then
				if need(2) then
					if b = 0 then
						err = 1
						errtxt = "modulo 0"
					elsif a = 0 then
						x = 0
						done(2)
					elsif integer(a) = 0 or integer(b) = 0 then
						err = 1
						errtxt = "imod (integer modulo) argument(s) not integer"
					elsif sign(a) != sign(b) then
						err = 1
						errtxt = "modulo arguments must have the same sign"
					else
						x = remainder(a, b)
						done(2)
					end if
				end if
			elsif equal(opc, "mod") then
				if need(2) then
					if b = 0 then
						err = 1
						errtxt = "modulo 0"
					elsif a = 0 then
						x = 0
						done(2)
					elsif sign(a) != sign(b) then
						err = 1
						errtxt = "modulo arguments must have the same sign"
					else
						x = remainder(a, b)
						done(2)
					end if
				end if
			elsif equal(opc, "bin") then
				if need(2) then
					if a = 0 or isnat(a) = 0 or isnat(b) = 0 then
						err = 1
						errtxt = "binomial coefficient argument(s) not positive integer"
					elsif b > a then
						err = 1
						errtxt = "binomial coefficient 2nd parameter > 1st"
					else
						x = 1
						for j = 1 to b do
							x = x * (a + 1 - j) / j
						end for
						done(2)
					end if
				end if
			elsif equal(opc, "^") or equal(opc, "root") then
				if need(2) then
					if equal(opc, "root") then
						if b = 0 then
							err = 1
							errtxt = "0th root"
						elsif a < 0 then
							err = 1
							errtxt = "root of negative number"
						else
							b = 1 / b
						end if
					end if
					if err = 0 then
						if b = 0 then
							x = 1
							done(2)
						elsif a = 0 then
							if b > 0 then
								x = 0
								done(2)
							else
								err = 1
								errtxt = "zero raised to negative power"
							end if
						elsif log(abs(a)) * b > 2303 then
								err = 1
								errtxt = "operator ^ result > 1e1000"
						elsif a > 0 then
							x = apower(a, b)
							done(2)
						else
							if integer(b) then
								x = apower(a, b)
								done(2)
							else
								err = 1
								errtxt = "negative number raised to non-integer power"
							end if
						end if
					end if
				end if
			elsif equal(opc, "log^") then
				if need(2) then
					if a <= 0 then
						err = 1
						errtxt = "logarithm of 0 or negative value"
					elsif b <= 1 then
						err = 1
						errtxt = "logarithm base 0, negative or <= 1"
					else
						x = log(a) / log(b)
						done(2)
					end if
				end if
			elsif equal(opc, "randint") then
				if need(2) then
					if integer(a) and integer(b) and b > a then
						x = rand_range(a, b)
						done(2)
					else
						err = 1
						errtxt = "random integer range error, arguments must be integer, a < b"
					end if
				end if
			elsif equal(opc, "randnd") then
				if need(2) then
					if a = 0 and b = 0 then b = 1 end if          -- 0 0 same as 0 1 (standard normal distribution)
					if b > 0 then
						x = ndrand(a, b)
						done(2)
					else
						err = 1
						errtxt = "standard deviation of normal distribution must be > 0"
					end if
				end if

			-- measurement unit conversion operators (2 arguments)

			elsif equal(opc, ":msdec") then
				if need(2) then
					if a < 0 then
						err = 1
						errtxt = "minutes < 0"
					elsif b < 0 then
						err = 1
						errtxt = "seconds < 0"
					else
						x = (a + (b / 60)) / 60
						done(2)
					end if
				end if

			-- statistical operators, variable argument number

			elsif equal(opc, "n") then
				first = 0
				x = ix - 1
				done(ix-1)
				restorehead()
			elsif equal(opc, "n+") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						if round0(stack[j]) > 0 then x += 1 end if
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "n0") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						if round0(stack[j]) = 0 then x += 1 end if
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "n-") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						if round0(stack[j]) < 0 then x += 1 end if
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "sum") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						x = x + stack[j]
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "mean") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						x = x + stack[j]
					end for
					x = x / (ix - 1)
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "hmean") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						if stack[j] <= 0 then
							err = 1
							errtxt = "negative or 0 argument not allowed for harmonic mean"
							exit
						else
							x = x + 1 / stack[j]
						end if
					end for
					if err = 0 then
						x = x / (ix-1)
						x = 1 / x
						done(ix-1)
						restorehead()
					end if
				end if
			elsif equal(opc, "gmean") then
				first = 0
				if need(1) then
					x = 1
					for j = 1 to ix-1 do
						if stack[j] <= 0 then
							err = 1
							errtxt = "negative or 0 argument not allowed for geometric mean"
							exit
						else
							x = x * stack[j]
						end if
					end for
					if err = 0 then
						x = apower(x, 1 / (ix-1))
						done(ix-1)
						restorehead()
					end if
				end if
			elsif equal(opc, "sqsum") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						x = x + stack[j] * stack[j]
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "rcpsum") then
				first = 0
				if need(1) then
					x = 0
					for j = 1 to ix-1 do
						if stack[j] = 0 then
							err = 1
							errtxt = "reciprocal value of 0"
							exit
						else
							x = x + 1 / stack[j]
						end if
					end for
					if err = 0 then
						done(ix-1)
						restorehead()
					end if
				end if
			elsif equal(opc, "prod") then
				first = 0
				if need(1) then
					x = 1
					for j = 1 to ix-1 do
						x = x * stack[j]
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "min") then
				first = 0
				if need(1) then
					x = stack[1]
					for j = 2 to ix-1 do
						if (stack[j]) < x then x = stack[j] end if
					end for
					done(ix-1)
				end if
			elsif equal(opc, "max") then
				first = 0
				if need(1) then
					x = stack[1]
					for j = 2 to ix-1 do
						if (stack[j]) > x then x = stack[j] end if
					end for
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "median") then
				first = 0
				object med
				med = {}
				if need(1) then
					for j = 1 to ix-1 do
						med = append(med, stack[j])
					end for
					med = sort(med)
					if even(ix) then
						x = med[ix/2]
					else
						a = (ix-1)/2
						b = (ix+1)/2
						x = (med[a] + med[b]) / 2
					end if
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "sdev") then
				first = 0
				if need(2) then
					y = 0
					for j = 1 to ix-1 do
						y = y + stack[j]
					end for
					y = y / (ix - 1)
					x = 0
					for j = 1 to ix-1 do
						x = x + (stack[j] - y) * (stack[j] - y)
					end for
					x = sqrt (x / (ix - 1))
					done(ix-1)
					restorehead()
				end if
			elsif equal(opc, "item") then
				first = 0
				if need(2) then
					y = stack[ix-1]
					if y = 0 then
						err = 1
						errtxt = "Item number is zero"
					elsif isnat(abs(y)) = 0 then
						err = 1
						errtxt = "Item number is not an integer"
					elsif abs(y) > ix-2 then
						err = 1
						errtxt = "Item number outside length of list"
					elsif y > 0 then
						x = stack[y]
						if y = ix-2 then loop = 0 end if                      -- referencing last item in list with positive index triggers loop exit
						done(ix-1)
						restorehead()
					else
						y = -y
						x = stack[ix - y - 1]
						if y = ix-2 then                                      -- referencing first item in list with negative index triggers loop exit
							if i_repeat > 1 then loop = 0 end if                   -- but not in first loop pass
						end if
						done(ix-1)
						restorehead()
					end if
				end if

			-- pseudo operators

			elsif equal(opc, "|") then
				if length(head) > 0 or ix = 1 or ix = length(stack) then
					err = 1
					errtxt = "invalid position of n-argument delimiter |"
				else
					head = stack[1..ix-1]
					stack = stack[ix+1..$]
					ix = 0
				end if

			elsif equal(opc, "whisk") then
				first = 0
				if need(2) then
					if ix < length(stack) then
						whisk_a = stack[ix-2]
						whisk_b = stack[ix-1]
						flag_whisk = 1
						stack = stack[1..ix-3] & stack[ix+1..$]
						ix -= 3
					end if
				end if

			elsif equal(opc, "a") or equal(opc, "b") then
				if flag_whisk = 1 then
					if equal(opc, "a") then
						stack[ix] = whisk_a
					else
						stack[ix] = whisk_b
					end if
				else
					err = 1
					errtxt = "A and B can only be used after WHISK"
				end if

			elsif equal(opc, "done") then
				first = 0
				if need(1) then
					stack[ix] = stack[ix-1]
					stack = stack[ix..$]
					ix = 1
				end if

			-- operator not recognized, err code 2 when first word in line, only used at the end of processline() to decide whether update hyin
			else
				if find(opc[1], "-.0123456789") then
					err = 1
					errtxt = stack[ix][2] & " is not a valid number"
				else
					if flag_fw then
						err = 2
						errtxt = "command " & stack[ix][2] & " not recognized"
					else
						err = 1
						errtxt = "term " & stack[ix][2] & " not recognized"
					end if
				end if

			end if

		end if
		if err then exit end if
		flag_fw = 0
	end while
	
	if err = 0 then
		if length(head) then
			err = 1
			errtxt = "no n-argument operator after delimiter |"
		else
			if length(stack) = 1 then
				result = stack[1]
			else
				err = 1
				errtxt = "argument count mismatch, remaining on stack: " & sprint(length(stack)-1)
			end if
		end if
	end if

end procedure