-- Hypatia
-- version 4.5

global procedure dispversion()
	disp("Hypatia version 4.5, (c) 2007-2025 by Robert Schaechter")
	disp("http://hypatia-rpn.net")
	disp("This software is offered 'as is.' No warranty of any kind is provided or implied.")
	disp("This is a Phix program, http://phix.x10.mx/")
	disp("")
end procedure


global procedure dispbasics()
		disp("For help use HELP command")
		disp("To see current settings, enter ?")
		disp("To see current variables, enter SHOW")
		disp("To quit, enter Q")
		disp("All input is case insensitive\n")
		disp("To open the user manual enter HELP doc\n")
end procedure


global procedure disphelp()
	object op
	if length(stack) > 1 then op = lower(stack[2]) else op = "" end if
	op = "." & op & "."
	disp("")
	if equal (op, "..") then
		dispversion()
		disp("Main help topics:")
		disp("HELP 1 -- operators with 1 argument")
		disp("HELP 2 -- operators with 2 arguments")
		disp("HELP n -- operators with n arguments")
		disp("HELP c -- constants")
		disp("HELP var -- variables")
		disp("HELP ude -- user defined elements")
		disp("HELP cmd -- commands")
		disp("HELP fmt -- output format commands")
		disp("HELP opt -- command line options")
		disp("HELP cs -- control symbols")
		disp("HELP other -- various topics")
		disp("HELP followed by specific operator or command -- detailed information")
		disp("HELP license")
		disp("HELP contact")
		disp("")
		disp("HELP ? -- list of all available help topics")
		disp("For current settings, enter ?")
		disp("To exit program, enter Q or QUIT, or close the console window")
		disp("")
		disp("All input is case insensitive, you can type upper case or lower case characters")
		disp("In help texts, upper case is used for operators or commands")
		disp("The input editor lets you use left and right arrows, home, end, backspace, and delete keys")
		disp("At leftmost cursor position up and down arrows let you scroll through past inputs")
		disp("ENTER inputs the line regardless of cursor position")
		disp("")
		disp("For more information ope the user manual with HELP doc")
	elsif equal(op, ".?.") then
		disp("List of all currently available help topics:")
		disp("HELP 1 2 n c (1, 2, n argument operators, constants)")
		disp("HELP cmd opt (commands, command line options)")
		disp("HELP x (various topics), cs (control symbols), # ## #: comments")
		disp("HELP var sto = prompt del show save ude del@ show@ save@ (variables, UDEs)")
		disp("HELP + - * / // \\ ^ log^ mod imod bin randint randnd round rounds gate uplimit lolimit abslimit")
		disp("HELP is0 isnot0 is+ is+0 is- is-0 equal unequal <> << <= >> >= multiple")
		disp("HELP % %+ %- %? %n %a")
		disp("HELP +- rcp abs sign int frac digits isposint isprime sq sqrt cube cbrt ln lg !")
		disp("HELP rad deg sin cos tan asin acos atan")
		disp("HELP pi e tau phi rand dup whisk done |")
		disp("HELP :f :c :mi :km :nautmi :nautkm :in :cm :ft :m :lb :kg :oz :g :toz :tg :gal :l :mpg :msd :rad :deg")
		disp("HELP $ $$ auto$")
		disp("HELP n n+ n0 n- sum mean gmean hmean median sdev sqsum rcpsum prod min max item")
		disp("HELP fmt fdec fhex fsci flakh fmillion fshort flong f'")
		disp("HELP mode use deg rad integer echo log copyall")
		disp("HELP run _ scripts files edit edin edini edlog ednotes exteditor cccommand")
		disp("HELP repeat do loop $loop maxloop i i1: i*: isloop passes time if then also else nothing endloop skip abort")
		disp("HELP zero $zero")
		disp("HELP & && hy buffer copy = == reset")
		disp("HELP () insert insert-files @ ude show@ del@ save@")
		disp("HELP license contact")

	elsif equal(op, ".doc.") then
		object docpath
		if length(hdir) then
			docpath = hdir & "doc\\index.htm"
		else
			docpath = ".\\doc\\index.htm"
		end if
		if get_file_type(docpath) = 1 then
			system(docpath)
		else
			disp("Error: file " & docpath & "not found")
			disp("You have to look for the folder with the documentation and open the file index.htm")
		end if
			
	elsif equal(op, ".1.") then
		disp("Operators with 1 argument:")
		disp("+- RCP ABS SIGN IS0 ISNOT0 IS+ IS+0 IS- IS-0 DIGITS ISPOSINT ISPRIME INT FRAC SQ SQRT CUBE CBRT !")
		disp("LN LOG10 LOG2 EXP EXP10 EXP2 RAD DEG SIN COS TAN ASIN ACOS ATAN")
		disp("Measurement unit conversion operators:")
		disp(":F :C :MI :KM :NAUTMI :NAUTKM :IN :CM :FT :M :LB :KG :OZ :G :TOZ :TG :GAL :L :MPG")
	elsif equal(op, ".2.") then
		disp("Operators with 2 arguments:")
		disp("+ - * / // \\ ^ ROOT LOG^ ROUND ROUNDS GATE UPLIMIT LOLIMIT ABSLIMIT MOD IMOD BIN RANDINT RANDND")
		disp("EQUAL UNEQUAL <> << <= >> >= MULTIPLE")
		disp("% %+ %- %? %n %a")
		disp("Measurement unit conversion operators:")
		disp(":MSDEC")
	elsif equal(op, ".cmd.") then
		disp("Commands:")
		disp("  For oputput format commands see HELP fmt,")
		disp("  for variable and user defined element commands see HELP var or HELP ude")
		disp("? -- display program version and current settings")
		disp("Q or QUIT -- exit program")
		disp("& -- clear result file hy, useful with REPEAT loop and accumulation mode")
		disp("&& -- begin new line in result file hy, useful with REPEAT loop and accumulation mode")
		disp("= -- show last calculation result (value of $) in currently specified format")
		disp("== -- show last calculation result (value of $) in decimal format with up to 18 digits")
		disp("HY -- show result file hy")
		disp("BUFFER START | SHOW | FLUSH -- start buffer mode, show buffer, write buffer to hy")
		disp("BUFFER SAVE filename [comment] -- save buffer to file, with optional comment line")
		disp("COPY -- copy result file hy to clipboard (not allowed in buffer mode, see also HELP copyall)")
		disp("COPIN -- copy last calculation input line to clipboard")
		disp("EDIT -- view or edit result file hy with external editor")
		disp("EDIT filename -- view, edit or create file with external editor")
		disp("EDIN -- view or edit file hyin, the most recent calculation input line")
		disp("EDINI -- view or edit configuration file hy.ini")
		disp("EDLOG -- view or edit log file hy.log")
		disp("EDNOTES -- view or edit notes file notes.txt")
		disp("EXTEDITOR executable filename -- set external editor other than default notepad.exe")
		disp("FILES -- lists all files in the Hypatia program folder")
		disp("RESET -- delete all variables and UDEs, restore default settings, run hy.ini")
		disp("RUN filename -- execute script file")
		disp("_filename -- same as RUN filename, but only shows end result")
		disp("REPEAT -- repeats last calculation input line, similar to RUN hyin")
		disp("REPEAT n -- loop command, repeats last calculation input line n times")
		disp("DO n :: ... -- loop command, executes calculation or script file n times")
		disp("DO * :: ... -- executes calculation or script the maximum number of times (100000 by default)")
		disp("*_filename -- short for DO * :: _filename")
		disp("MAXLOOP n -- sets maximum number of passes for loop commands")

	elsif equal(op, ".opt.") then
		disp("Command line options:")
		disp("-c   i/o files location is current folder (default: program folder)")
		disp("-d   set angle unit to degrees (default: radians)")
		disp("-e   set echo mode on (default: off)")
		disp("-i   do not execute ini file (default: execute hy.ini)")
		disp("-l   log input and output to file hy.log (default: no logging)")
		disp("-r filename -- run script file")

	elsif match(op, ".mode.use.deg.rad.integer.echo.log.copyall.") then
		disp("Mode commands:")
		disp("USE DEG | RAD     use degrees or radians as angle units for trigonometric functions")
		disp("INTEGER ON | OFF  enable (disable) integer bias, default = ON")
		disp("AUTO$ ON | OFF    enable (disable) automatic insertion of $, default = ON (see HELP $)")
		disp("ECHO ON | OFF     display (don't display) executed files, default = OFF")
		disp("LOG ON | OFF      log (do not log) input and output to file hy.log, default = OFF")
		disp("COPYALL ON | OFF  copy (don't copy) all results to clipboard (ignored in scripts and buffer mode)")
		disp("")
		disp("All the above mode commands return the present status if given without arguments")

	elsif match(op, ".var.sto.=.prompt.del.show.save.ude.del@.show@.save@.") then
		disp("Names of variables must begin with $")
		disp("STO $var       assigns value of last result to variable (creates variable if it doesn't exist)")
		disp("$var = ...     assigns value or result of expression to variable")
		disp("               (creates variable if it doesn't exist, does not update $ and hy, updates hyin)")
		disp("PROMPT $var [comment]   prompts user for value of variable, can be number or expression")
		disp("               (creates variable if it doesn't exist, does not update $ and hy,")
		disp("               does not update hyin, empty input leaves variable unchanged)")
		disp("               comment text will be displayed")
		disp("DEL $var       deletes variable")
		disp("SHOW           displays variable names and values")
		disp("               If angle unit is set to degrees, this will be displayed (radians is default)")
		disp("SHOW $var1 $var2 ...      displays these variables (can include loop index I)")
		disp("")
		disp("Names of user defined elements (UDEs) must begin with @")
		disp("@ude = ...     assigns text (numbers and/or operators) to user defined element")
		disp("DEL@ @ude      deletes user defined element")
		disp("SHOW@          displays user defined elements")
		disp("SHOW@ @ude1 @ude2 ...     displays these user defined elements")
		disp("")
		disp("Variables and UDEs are lost on exit from Hypatia, unless they are saved")
		disp("SAVE filename  saves all variables to specified file (no file extension added by default!)")
		disp("SAVE@ filename saves all UDEs to specified file (no file extension added by default!)")
		disp("               filename cannot include spaces, and will always be lower case")
		disp("               anything written after filename gets written to the file as a comment line")
		disp("_filename      loads all variables or UDEs saved to that file (see HELP run)")
		disp("SHOW, DEL and SAVE display, delete and save variables, SHOW@, DEL@, SAVE@ user defined elements!")
		disp("For the command = see HELP ==")

	elsif match(op, ".fmt.fdec.fhex.fsci.flakh.fmillion.fshort.flong.f'.") then
		disp("Output format commands:")
		disp("FDEC         show results in decimal format (scientific format for very large or small numbers)")
		disp("FDEC n       show results with n digits (max. 15) after decimal point")
		disp("FDEC -n      show results with n digits (incl. leading zeros) before decimal point (max. 15)")
		disp("FHEX         show results in hexadecimal binary format")
		disp("FHEX n       show results in hexadecimal format with n digits or more (max. 12)")
		disp("FBIN         show results in binary format")
		disp("FBIN n       show results in binary format with n digits or more (max. 48)")
		disp("FSCI         show results in scientific format")
		disp("FLAKH        show numbers larger/equal 1e5 in lakh, larger/equal 1e7 in crore")
		disp("FMILLION     show numbers larger/equal 1e6 in million")
		disp("FSHORT       show results with up to 9 decimal digits")
		disp("FLONG        show results with up to 15 decimal digits (default)")
		disp("F' ON | OFF  apostrophe format, show large numbers structured with with apostrophes")
		disp("               decimal numbers every 3 digits before the decimal point,")
		disp("               hexadecimal numbers every 4 digits, binary numbers every 8 digits")
		disp("After an output format command the last result is shown in the newly chosen format,")
		disp("but the result file hy is not updated. To update hy, enter $")
		disp("$ on its own is a command. Within a script it does nothing")
		disp("$ & and $ && are calculations, they also work in scripts")

	elsif match(op, ".other.cs.#.#:.##.comments.") then
		disp("Various topics:")
		disp("Variable names must start with $")
		disp("$ is last result, $$ is last result but one - see HELP $")
		disp("File hy has the last result, in the currently chosen format")
		disp("Command HY shows content of result file hy, command COPY copies it to the clipboard")
		disp("Command & clears file hy, command && begins new line in hy")
		disp("File hyin always has the most recent calculation input line")
		disp("(file) includes content of file in input line - see HELP ()")
		disp("")
		disp("Control symbols:")
		disp("Accumulation mode: end calculation line with & or && to accumulate results in file hy")
		disp("... & separates results by spaces, ... && adds the new result as a new line")
		disp("Silent mode: end input line with # to prevent result being written to file hy")
		disp("Debugging mode: end input line with ? to display intermediate results of calculation")
		disp("")
		disp("Comments:")
		disp("# begins comment line (only useful with log file, or in scripts)")
		disp("#: begins comment line in a script that will be displayed at each pass of a loop")
		disp("## begins comment line in a script that will not be displayed")
		disp("... ## comment -- comments can be be written after commands or calculations in scripts")
		disp("")
		disp("For constants, pseudo constants and pseudo operators, see HELP c")
		disp("For scripts, see HELP scripts, for loops, see HELP loop, for buffer, see HELP buffer")
		disp("Command line calculation mode: hy [options] expression to be calculated")

	elsif match(op, ".c.pi.tau.e.phi.rand.dup.whisk.done.|.") then
		disp("Constants, to be used like numbers:")
		disp("E       2.718281828459045235")
		disp("PI      3.141592653589793238")
		disp("TAU     6.28318530717958648 (2 PI)")
		disp("PHI     1.618033988749894848 (the golden ratio)")
		disp("Pseudo constants, so called because they function like constants in Hypatia syntax:")
		disp("RAND    random number, range 0 to 1 (do not confuse with RANDINT and RANDND operators)")
		disp("DUP     duplicates the preceding number, constant or variable in the input line")
		disp("        Example:")
		disp("        ? 6 sq dup +")
		disp("        = 42")
		disp("I       loop index (in REPEAT loops loop index +1), always 1 when not in a loop")
		disp("ISLOOP  1 if script runs in a loop, 0 if script was called directly")
		disp("TIME    time in seconds after start of loop, outside of loops always 0")
		disp("")
		disp("Pseudo operators:")
		disp("WHISK   deletes (whisks away) the two values left of itself, and saves them")
		disp("        under the names A and B (only within the current calculation line)")
		disp("        Example:")
		disp("        ? 3 4 WHISK A SQ B SQ + SQRT")
		disp("        = 5")
		disp("DONE    deletes anything but the value immediately left of itself from the stack")
		disp("        for instance useful for calculating generalized Fibonacci sequences")
		disp("|       n-argument delimiter, hides anything to its left from the next n-argument operator")
		disp("        Example:")
		disp("        ? 20 | 3 5 8 SUM -")
		disp("        = 4")

	elsif match(op, ".$.$$.auto$.") then
		disp("All variable names start with $ -- see HELP var")
		disp("Variables $ and $$ have special meanings:")
		disp("$   result of the last successful calculation")
		disp("$$  result of the last but one successful calculation")
		disp("    Example:")
		disp("    2 3 +")
		disp("    = 5")
		disp("    8 $ -")
		disp("    = 3")
		disp("    $$ $ ^")
		disp("    = 125")
		disp("")
		disp("By default $ can be omitted at the beginning of the input line, for operators requiring")
		disp("one or two arguments. Example:")
		disp("    2 3 +")
		disp("    = 5")
		disp("    7 *")
		disp("    = 35")
		disp("This default behavior can be disabled with the command AUTO$ OFF")
		disp("AUTO$ ON enables it again, AUTO$ shows current status")
		disp("")
		disp("After each calculation, the value of $ is saved in the file hy -- see HELP hy")
		disp("At the next start of Hypatia, it can be inserted into the input line as (hy)")
		disp("For information on (), see HELP ()")
		disp("")
		disp("$ on its own is a command: write the value of $ to hy in the currently chosen format")
		disp("  Not allowed in scripts or in buffer mode. $ & and $ &&, though, are calculation lines")

	elsif match(op, ".&.&&.hy.copy.==.reset.") then
		disp("&      clears result file hy")
		disp("&&     begins new line in file hy")
		disp("  Do not confuse commands & and && with accumulation mode control symbols & and &&, see HELP cs")
		disp("  Accumulation mode: end calculation line with & or && to accumulate results in file hy")
		disp("  ... & separates results by spaces, ... && adds the new result as a new line")
		disp("HY     shows result file hy, useful in connection with accumulation mode")
		disp("... (hy) ...   inserts content of hy in input line")
		disp("For BUFFER commands see HELP buffer")
		disp("")		
		disp("COPY   copies hy to the clipboard")
		disp("EDIT   lets you view and edit file hy (see HELP edit)")
		disp("")
		disp("=      shows last result (value of $) in currently chosen format")
		disp("       (do not confuse the command = with = in assignment commands, see HELP var")
		disp("==     shows last result (value of $) in decimal format with up to 18 digits")
		disp("       (do not confuse the command == with the == compare operator, same as EQUAL)")
		disp("")
		disp("RESET  deletes all variables and UDEs, resets options to default or to command line parameters,")
		disp("       runs hy.ini, unless Hypatia was started with the -i option")
		disp("       the old name of this command was INIT, which still works")

	elsif match(op, ".buffer.") then
		disp("BUFFER START    clears buffer, sets buffer mode to ON")
		disp("                all results will be written to buffer, not to result file hy")
		disp("BUFFER SHOW     shows content of buffer")
		disp("BUFFER SAVE filename    writes content of buffer to specified file")
		disp("                (anything written after the file name gets written to the file as a comment line)")
		disp("BUFFER FLUSH    writes buffer to file hy, overwrites hy unless buffer is empty,")
		disp("                clears buffer, sets buffer mode to OFF")
		disp("BUFFER DISCARD  deletes buffer, sets buffer mode to OFF")
		disp("BUFFER          shows buffer mode ON, ON but empty, or OFF")
		disp("... (buffer) ...     analogous to (hy), inserts content of buffer in input line")
		disp("")
		disp("When buffer mode is ON, accumulation mode controls ... & and ... && and commands & and &&")
		disp("work with the buffer exactly as they would otherwise work with hy")

	elsif match(op, ".run._.script.scripts.files.") then
		disp("RUN filename   executes script, each line in the file is treated as an input line")
		disp("    The file hy.ini is processed as a script file at each start of Hypatia,")
		disp("    you can edit it according to your needs.")
		disp("    Script files may not contain RUN, RESET, REPEAT and DO commands")
		disp("    RUN command line is the only command line that gets written to hyin")
		disp("_filename      same as RUN filename, but only the final result is shown")
		disp("    Note: no space between _ and filename!")
		disp("(filename)     The content of the file is inserted into the input line at this point")
		disp("               see also HELP ()")
		disp("FILES          displays list of all files in the Hypatia program folder")
		disp("")
		disp("Note the difference between RUN filename and (filename):")
		disp("    RUN expects a file in which each line is a valid input line.")
		disp("    (filename) inserts the content of the file into the current input line,")
		disp("    it can contain numbers and/or operators, which can be separated by spaces or line breaks.")
		disp("All files will be read from and written to Hypatia's program folder (unless you use the -c option)")

	elsif match(op, ".edit.edin.edini.edlog.ednotes.exteditor.") then
		disp("FILES          displays list of all files in the Hypatia program folder")
		disp("EDIT filename  opens external editor to view or edit a file")
		disp("EDIT           view or edit result file hy with external editor")
		disp("EDIT filename  view, edit or create file with external editor")
		disp("               If file does not exist, you will be asked whether you want to create it")
		disp("               By default, the editor is Notepad, but you can change this with EXTEDITOR")
		disp("               File names can not have spaces or parentheses, and will always be lower case")
		disp("EDIN           view or edit file hyin, the most recent calculation input line")
		disp("EDINI          view or edit configuration file hy.ini")
		disp("EDLOG          view or edit log file hy.log")
		disp("EDNOTES        view, edit or create notes file notes.txt")
		disp("")
		disp("EXTEDITOR executable filename -- set external editor other than default notepad.exe")
		disp("  Filename can include a path, and can contain spaces and parentheses")
		disp("  Even if it contains spaces, do NOT enclose filename in quotation marks!")
		disp("  Without file name, the current external editor is shown")

	elsif match(op, ".repeat.do.loop.loops.maxloop.$loop.i.i1:.i*:.isloop.passes.time.if.then.also.else.also:.else:.nothing.endloop.skip.abort.") then
		disp("REPEAT         repeats last calculation input line, same as RUN hyin")
		disp("  If the line contains $ and/or $$, these values will be updated at each execution")
		disp("  If the line contains () insert files, their current content will be inserted")
		disp("  You can use EDIN to edit the recent input line and REPEAT to perform the edited calculation")
		disp("REPEAT n       loop command: repeats last calculation or last RUN command n times (default max. 99999")
		disp("DO n :: ...    loop command: executes calculation or script file n times (default max. 100000")
		disp("DO * :: ...    n is set to maximum (default is 100000, see MAXLOOP n")
		disp("*_filename     short for DO * :: _filename (run script max. number of times, or until exit condition")
		disp("               Loop commands can not be used within scripts")
		disp("               To exit the loop before n is reached set variable $loop to 0")
		disp("               To see all results, not just the final one, add ? after the number of passes")
		disp("$loop          To exit the loop before n is reached, set variable $loop to 0")
		disp("MAXLOOP n      Sets max. number of passes in loops to n (default 100000, max. 10 million")
		disp("I              in DO loop: loop index, in REPEAT n loop: loop index + 1, outside of loop: always 1")
		disp("I1: ...        as prefix to a script line: line gets executed when I = 1")
		disp("I*: ...        as prefix to a script line: line gets executed at the end of the loop")
		disp("ISLOOP         used in scripts: 1 if script runs in a loop, 0 if script was called directly")
		disp("PASSES         number of loop passes with which a script has been called")
		disp("TIME           time in seconds after start of loop, outside of loops always 0")
		disp("")
		disp("IF ... THEN    conditional execution of a command or calculation, or displays comment")
		disp("               IF variable or calculation: 0 = false, not 0 = true (zero threshold does NOT apply)")
		disp("ALSO: ...      executes command or calculation or shows comment, after IF condition is true")
		disp("ELSE: ...      executes command or calculation or shows comment, after IF condition is false")
		disp("               ELSE: inverts result of IF condition, which also applies to subsequent ALSO: lines")
		disp("  ALSO: IF ... and ELSE: IF ... are permitted")
		disp("NOTHING        does nothing, can in some cases be useful after IF ... THEN")
		disp("IF ... THEN PROMPT $loop  prompts for value of $loop, exits loop when $loop set to 0")
		disp("IF ... THEN SKIP          skips remaining lines in a script")
		disp("IF ... THEN ENDLOOP       exits loop after last line in script, same as setting $loop = 0")
		disp("IF ... THEN ABORT         combines ENDLOOP and SKIP, skips remaining lines in script, then exits loop")
		disp("            anything written after ENDLOOP or ABORT will be displayed as a comment")

	elsif match(op, ".zero.$zero.") then
		disp("$zero = ...    set variable $zero to any positive value, to change zero threshold from default 1e-16")
		disp("$zero = 0      set variable $zero to 0 to disable the zero threshold (only exactly 0 is zero)")
		disp("DEL $zero      deleting $zero variable returns zero threshold to its 1e-16 default")
		disp("")
		disp("               zero threshold applies to SIGN, IS0, ISNOT0, IS+, IS0+, IS-, IS0-,")
		disp("               the compare operators, 0 GATE, and to N+, N0, N-")
		disp("               zero threshold does not apply to IF (only false when exactly 0),")
		disp("               nor does it apply to // (always uses 1e-16)")
		disp("               0 GATE sets gate threshold to zero threshold, if that is 0 then to 1e-16")
		disp("               SIN and COS always round results less than 1e-18 to zero")

	elsif match(op, ".(.().insert.insert-files.") then
		disp("The content of a file can be inserted into the input line by putting the file name in ()")
		disp("This has two main purposes:")
		disp("1. to create your own operators:")
		disp("For example, to calculate the area of a circle from its diameter you need the expression")
		disp("2 / SQ PI * -- if you save 2 / SQ PI * as a file with the name circle, then you can calculate")
		disp("the the area of a circle by writing (circle) after the diameter:")
		disp("25 (circle) 5 ROUNDS")
		disp("= 490.87 (rounded to 5 significant digits)")
		disp("2. to input data for statistical operators. Example:")
		disp("(mydata.dat) MEAN")
		disp("(mydata.dat) SDEV")
		disp("Elements in the included file can be separated by spaces or put in separate lines, or both.")
		disp("All files will be read from and written to Hypatia's program folder (unless you use the -c option),")
		disp("any path that might prepend the file name will be relative to this folder.")
		disp("No file extension is assumed or added by default, you have to include it with the file name.")
		disp("A list of files in Hypatia's program folder is displayed with the FILES command.")
		disp("\nSee HELP ude for user defined elements!")

	elsif match(op, ".@.ude.show@.del@.save@.") then
		disp("User defined elements (UDEs) work similar to insert files -- see HELP insert --")
		disp("but they are stored internally, similar to variables.")
		disp("A UDE has a name that starts with @, and, like an insert file, a content which is a text")
		disp("that can consist of numbers, variable names and operators.")
		disp("1. You can use UDEs to create your own operators:")
		disp("For example, to calculate the area of a circle from its diameter,")
		disp("you can assign the expression 2 / SQ PI * to a user defined element, and name it @circle:")
		disp("@circle = 2 / SQ PI *")
		disp("If you add this line to hy.ini, then you have this UDE always available as an operator:")
		disp("25 @circle 5 ROUNDS")
		disp("= 490.87 (rounded to 5 significant digits)")
		disp("2. You can use an UDE for data, by reading it from an insert file:")
		disp("@mydata = (mydata.dat)")
		disp("or you can also directly assign a series of values to an UDE, for instance:")
		disp("@mydata = 3 8 4 9 5 2 7 11 6")
		disp("and then use n-argument operators on the data:")
		disp("@mydata MEAN")
		disp("@mydata SDEV")
		disp("")
		disp("SHOW@           displays user defined element names and values")
		disp("SHOW@ @ude1 @ude2 ...    displays these user defined elements")
		disp("DEL@ @ude       deletes user defined element")
		disp("SAVE@ filename  saves all UDEs to specified file (no file extension added by default!)")
		disp("                filename cannot include spaces, and will always be lower case")
		disp("                Anything written after the file name gets written to the file as a comment line")
		disp("_filename       loads all user defined elements saved to that file (see HELP run),")
		disp("                for UDEs which you need often it makes sense to add them to hy.ini")
		disp("SHOW, DEL and SAVE display, delete and save variables, SHOW@, DEL@, SAVE@ user defined elements!")

	elsif match(op, ".+-.rcp.abs.sign.is0.isnot0.is+.is+0.is0+.is-.is-0.is0-.int.frac.digits.isposint.isprime.sq.sqrt.cube.cbrt.ln.log10.log2.exp.exp10.exp2.!.") then
		disp("a +-       changes the algebraic sign of a")
		disp("a RCP      reciprocal value of a")
		disp("a ABS      absolute value of a")
		disp("a SIGN     1 if a is positive, 0 if zero, -1 if a is negative (zero threshold applies)")
		disp("a IS0      1 if a is zero, 0 if non-zero (zero threshold applies to all IS operators)")
		disp("a ISNOT0   1 if a is positive or negative, 0 if zero")
		disp("a IS+      1 if a is positive, 0 if zero or negative")
		disp("a IS+0     1 if a is positive or 0, 0 if negative (you can also write IS0+)")
		disp("a IS-      1 if a is negative, 0 if zero or positive")
		disp("a IS-0     1 if a is negative or 0, 1 if positive (you can also write IS0-)")
		disp("           for SIGN and the above IS operators the zero threshold applies, see HELP zero")
		disp("a ISPOSINT 1 if argument is positive integer, 0 if not (a must be < 1e18)")
		disp("a ISPRIME  1 if a is prime number, 0 if not (a must be 0 or positive integer)")
		disp("a DIGITS   number of digits before decimal point, if negative: number of zeros after dec. point")
		disp("a INT      truncated integer value of a (negative when a is negative), a must be < 1e18")
		disp("a FRAC     fractional part of a (negative when a is negative), a must be < 1e15")
		disp("a SQ       square of a")
		disp("a CUBE     cube of a")
		disp("a SQRT     square root of a (a must not be negative)")
		disp("a CBRT     cube root of a (a can be negative)")
		disp("a LN       natural logarithm (logarithm base e) of a")
		disp("a LOG10    logarithm base 10 of a")
		disp("a LOG2     logarithm base 2 of a")
		disp("a EXP      e to the power of a")
		disp("a EXP10    10 to the power of a")
		disp("a EXP2     2 to the power of a")
		disp("a !        factorial of a (a must be 0 or positive integer not larger than 449, 0 ! is 1 by definition")

	elsif match(op, ".rad.deg.sin.cos.tan.asin.acos.atan.") then
		disp("a RAD     converts angle from degrees to radians (do not confuse with USERAD command)")
		disp("a DEG     converts angle from radians to degrees (do not confuse with USEDEG command)")
		disp("  by default, all angles are in radians, you can change this using the USEDEG command")
		disp("a SIN     sine(a)")
		disp("a COS     cosine(a)")
		disp("          SIN and COS round results < 1e-18 to 0")
		disp("a TAN     tangent(a), abs. value of argument must be less than p/2 or 90 degrees")
		disp("a ASIN    arcsine(a)")
		disp("a ACOS    arccosine(a)")
		disp("a ATAN    arctangent(a)")

	elsif match(op, ".:f.:c.:mi.:km.:nautmi.:nautkm.:in.:cm.:ft.:m.:lb.:kg.:oz.:g.:toz.:tg.:gal.:l.:mpg.:msd.") then
		disp("a :F        converts degrees Centigrade to Fahrenheit")
		disp("a :C        converts degreed Fahrenheit to Centigrade")
		disp("a :MI       converts kilometers to international miles")
		disp("a :KM       converts international miles to kilometers")
		disp("a :NAUTMI   converts kilometers to nautical miles")
		disp("a :NAUTKM   converts nautical miles to kilometers")
		disp("a :IN       converts centimeters to inches")
		disp("a :CM       converts inches to centimeters")
		disp("a :FT       converts meters to feet (1 ft = 12 in)")
		disp("a :M        converts feet to meters (1 ft = 12 in)")
		disp("a :LB       converts kilograms to avoirdupois pounds")
		disp("a :KG       converts avoirdupois pouns to kilograms")
		disp("a :OZ       converts grams to avoirdupois ounces (1 lb = 16 oz)")
		disp("a :G        converts avoirdupois ounces to grams (1 lb = 16 oz)")
		disp("a :TOZ      converts grams to troy ounces")
		disp("a :TG       converts troy ounces to grams")
		disp("a :GAL      converts liters to US liquid gallons")
		disp("a :L        converts US liquid gallons to liters")
		disp("a :MPG      converts liters/km to miles/gallon, and vice versa")
		disp("a b :MSDEC  converts a minutes b seconds to decimal hours or degrees")
		disp("            (this is the only conversion operator that takes 2 arguments)")
		disp("To convert square measures, use length conversion twice, e.g. 65 :ft :ft")

	elsif match(op, ".+.-.*././/.\\.^.root.log^.mod.imod.bin.randint.randnd.round.rounds.gate.uplimit.lolimit.abslimit.") then
		disp("a b +         a plus b")
		disp("a b -         a minus b")
		disp("a b *         a times b")
		disp("a b /         a divided by b")
		disp("a b //        a minus b divided by b (a if b=0), set to 0 if < 1e-16 (useful as loop exit condition)")
		disp("a b \\         a divided by b, absolute value rounded off to next integer")
		disp("a b ^         a to power of b")
		disp("a b ROOT      bth root of a")
		disp("a b LOG^      logarithm of a base b (a must be positive, b must be > 1)")
		disp("a b MOD       remainder, a modulo b (a and b must have the same sign)")
		disp("a b IMOD      remainder, a modulo b (a and b must be integer and have the same sign)")
		disp("a b BIN       binomial coefficient")
		disp("              a must be positive integer, b must be 0 or positive integer, a 0 BIN is 1 by definition")
		disp("a b RANDINT   random integer, within range of a to b (do not confuse with RAND pseudo constant)")
		disp("a b RANDND    normal distributed random number, mean = a, sigma = b (limited to 6 sigma)")
		disp("              0 0 RANDND uses standard normal distribution, same as 0 1 RANDND")
		disp("a b ROUND     round a to b digits after decimal point")
		disp("              if b is negative, then round to b zeros before decimal point")
		disp("              b must be an integer in the range of -9 to 15")
		disp("a b ROUNDS    round a to b significant digits, b must be an integer in the range of 1 to 17")
		disp("a b GATE      zero if absolute value of a is lower than b, otherwise a")
		disp("              if b = 0 gate threshold is set zero threshold, if that is 0 then to 1e-16")
		disp("a b UPLIMIT   b is upper limit for a (result is a, but not higher than b)")
		disp("a b LOLIMIT   b is lower limit for a (result is a, but not lower than b)")
		disp("a b ABSLIMIT  b is limit for absolute value of a (result is a, but not higher than b or lower than -b)")

	elsif match(op, ".equal.unequal.<>.<<.<=.>>.>=.multiple.") then
		disp("a b EQUAL     1 if a and b are equal, 0 if not (zero threshold applies)")
		disp("a b UNEQUAL   1 if a and b are unequal, 1 if equal (zero threshold applies)")
		disp("    == <>     same as EQUAL and UNEQUAL (do not confuse == operator with == command)")
		disp("a b <<        1 if a < b, 0 if not (zero threshold applies to all compare operators, see HELP zero)")
		disp("a b <=        1 if a <= b, 0 if not")
		disp("a b >>        1 if a > b, 0 if not")
		disp("a b >=        1 if a >= b, 0 if not")
		disp("a b MULTIPLE  a / b if a is positive multiple of b, 0 if not (a and b must be integer)")

	elsif match(op, ".%.%+.%-.%?.%n.%a.") then
		disp("a b %     b percent of a")
		disp("a b %+    a increased by b percent")
		disp("a b %-    a diminshed by b percent")
		disp("a b %?    how many percent of a is b?")
		disp("a b %N    net: if a is net amount plus b percent, what is the net amount?")
		disp("a b %A    agio: if a is gross amount including b percent, what is the agio amount?")

	elsif match(op, ".n.n+.n0.n-.sum.mean.gmean.hmean.median.sdev.sqsum.rcpsum.prod.min.max.item.") then
		disp("Operators with n arguments (statistical operators)")
		disp("These operators process all arguments to their left")
		disp("Arguments can be read from a file using (filename) - see HELP ()")
		disp("a ... n    N       number of arguments")
		disp("a ... n    N+      number of positive values")
		disp("a ... n    N0      number of zero values")
		disp("a ... n    N-      number of negative values")
		disp("                   zero threshold applies to N+, N0 and N-, see HELP zero")
		disp("a ... n    SUM     total sum")
		disp("a ... n    MEAN    arithmetic mean")
		disp("a ... n    GMEAN   geometric mean (all arguments must be > 0)")
		disp("a ... n    HMEAN   harmonic mean (all arguments must be > 0)")
		disp("a ... n    MEDIAN  median")
		disp("a ... n    SDEV    standard deviation")
		disp("a ... n    SQSUM   sum of squares")
		disp("a ... n    RCPSUM  sum of reciprocals")
		disp("a ... n    PROD    product of all arguments")
		disp("a ... n    MIN     minimum value")
		disp("a ... n    MAX     maximum value")
		disp("a ... n i  ITEM    value of item i in list of arguments")
		disp("                   if i is negative, counts backwards from end of list")
		disp("                   loop ends after last item in list was referenced (first item when i is negative)")
		disp("All statistical operators require at least 1 argument,")
		disp("except for SDEV and ITEM which require 2, and N which can have 0")

	elsif match(op, ".license.licence.contact.") then
		dispversion()
		disp("No legalese, just very simple rules.")
		disp("You may do with the executable file, the documentation, and the source code whatever you want,")
		disp("under three conditions:")
		disp("- that you credit the original author (see above), and")
		disp("- that you document the changes you have made, at least in general terms, and that")
		disp("- under no circumstances you are allowed to modify this program in a way that could be potentially")
		disp("harmful to the user, or to distribute it in a potentially harmful way, for instance with an installer")
		disp("that installs other software or makes undocumented or unrequested changes to the user's computer.")
		disp("")
		disp("The executable, the source code and the documentation are offered 'as is.'")
		disp("No warranty of any kind is provided or implied. All support is given voluntarily.")
		disp("")
		disp("If you improve or otherwise modify the program, port it to Linux etc., please let me know.")
		disp("I'd also like to hear if you find the program useful, or if you have any suggestions or complaints.")
		disp("Important: please tell me if you've found or suspect a bug, if you found an error in the documentation,")
		disp("or if you find parts of the documentation to be unclear, incomplete, or difficult to understand.")
		disp("")
		disp("Contact: Robert Schaechter, Vienna/Austria, robert@drs.at")
		disp("Please report bugs. Feel free to ask questions or make suggestions.")
		disp("Please put Hypatia in the subject line of your message")
		
	else
		disp("Help topic " & stack[2] & " not available")
		disp("see HELP ? for a list of all available help topics")
	end if
	disp("")
end procedure
